/* 
 * Copyright (C) 2006, Intel Corporation
 * 
 * This file is part of the Linux-ready Firmware Developer Kit
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation;version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program in a file named COPYING; if not, write to the 
 * Free Software Foundation, Inc., 
 * 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA
 */

#define _GNU_SOURCE 
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdint.h>
#include <errno.h>

#include <glib.h>

#include "biostest.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/klog.h>

#undef OVERRIDE_E820


/* 
 * infrastructure for modules to find simple acpi tables 
 */


unsigned long RSDP_ADDRESS;  /* filled in by the dmesg parser in e820.c */

static unsigned long rsdt;
static unsigned long xsdt;
static int rsdt_size;


struct rsdp_table {
	char		signature[8];
	unsigned char	checksum;
	char		oem_id[6];
	unsigned char	revision;
	unsigned int	rsdt_address;
	unsigned int	rsdt_length;
	uint64_t	xsdt_address;
	unsigned char	checksum2;
} __attribute__((packed));

struct generic_acpi_header {
	char	signature[4];
	unsigned int	length;
	unsigned char	revision;
	unsigned char	checksum;
	unsigned char	oem_id[6];
	unsigned char	oem_table_id[8];
	unsigned int	oem_revision;
	unsigned int	creator_id;
	unsigned int	creator_rev;
} __attribute__((packed));



#define PAGE_MASK (~(getpagesize()-1))
#define PAGE_OFFSET_MASK (getpagesize()-1)


static unsigned char checksum(unsigned char *ptr, int size)
{
	unsigned char sum = 0;
	int i;
	for (i = 0; i<size; i++)
		sum += ptr[i];
	return sum;
}

struct generic_acpi_header *copy_acpi_header(uint64_t address, char *name)
{
	int header_crosses_page = 0;
	int residu;
	int fd;
	struct generic_acpi_header *header;
	char *table_page;
	int count;

	if (!e820_is_reserved(address)) {
		char buffer[512];
		sprintf(buffer, "E820: %s (0x%lx) is not in reserved or ACPI memory!", name, (unsigned long)address);
		report_result(NULL, FAIL, buffer, NULL, "e820://");
#ifndef OVERRIDE_E820
		return NULL;
#endif
	}

	/* Step 1: mmap and copy only the header */

	/* check if the header crosses page boundary, because if it does we need to
	   mmap 2 pages */

	if ( (address & PAGE_MASK) != ( (address+36) & PAGE_MASK))
		header_crosses_page = 1;

	residu = address & PAGE_OFFSET_MASK;

	fd = open("/dev/mem", O_RDONLY);
	if (fd<=0)
		return NULL;

	count = getpagesize();
	if (header_crosses_page)
		count += getpagesize();

	table_page = mmap(NULL, count , PROT_READ, MAP_SHARED, fd, address & PAGE_MASK);

	if (!table_page)
		return NULL;

	if (table_page==(void*)-1) {
		printf("Error is %s \n", strerror(errno));
		return NULL;
	}

	header = malloc(36);
	assert(header!=NULL);
	memcpy(header, table_page + residu, 36);
	munmap(table_page, count);

	if (header->length < 36) {
		char buffer[512];
		char buf[5];
		buf[4] = 0;
		memcpy(buf, header->signature, 4);
		sprintf(buffer, "ACPI table %s has an invalid length of %i", buf, header->length);
		report_result(NULL, FAIL, buffer, NULL, NULL);
		free(header);
		return NULL;
	}

	return header;
}



char *copy_acpi_table(uint64_t address, char *name)
{
	int residu;
	int fd = 0;
	struct generic_acpi_header *header = NULL;
	char *table_page;
	char *newtable;
	int count;

	if (!e820_is_reserved(address)) {
		char buffer[512];
		sprintf(buffer, "E820: %s (0x%lx) is not in reserved or ACPI memory!", name, (unsigned long)address);
		report_result(NULL, FAIL, buffer, NULL, "e820://");
#ifndef OVERRIDE_E820
		return NULL;
#endif
	}


	header = copy_acpi_header(address, name);

	if (header == NULL)
		return NULL;

	/* Step 2: validate the header */

	if (strncmp(header->signature,name,4)!=0) {
		char buffer[1024];
		char buf[5];
		buf[4] = 0;
		memcpy(buf, header->signature, 4);
		sprintf(buffer, "ACPI table: expected '%s', but got '%s'", name, buf);
		report_result(NULL, WARN, buffer, NULL, NULL);
		goto out;
	}

	/* Step 3: mmap and copy the whole table */

	newtable = malloc(header->length);
	assert(newtable != NULL);

	residu = address & PAGE_OFFSET_MASK;
	count = residu + header->length;
	if (count & PAGE_OFFSET_MASK) 
		count = (count & PAGE_MASK) + getpagesize();
	
	fd = open("/dev/mem", O_RDONLY);
	if (fd<=0)
		goto out;

	table_page = mmap(NULL, count , PROT_READ, MAP_SHARED, fd, address & PAGE_MASK);

	if (!table_page)
		goto out;

	if (table_page==(void*)-1) {
		printf("Error is %s \n", strerror(errno));
		goto out;
	}

	memcpy(newtable, table_page + residu, header->length);

	munmap(table_page, count);
	close(fd);	

	if (checksum((unsigned char*)newtable, header->length) != 0) {
		char buffer[1024];
		char buf[5];
		buf[4] = 0;
		memcpy(buf, header->signature, 4);
		sprintf(buffer, "ACPI table %s has incorrect checksum (%i)", buf, checksum((unsigned char*)newtable, header->length));
		report_result(NULL, FAIL, buffer, NULL, NULL);
	}

	free(header);
	return newtable;

out:
	free(header);
	if (fd)
		close(fd);
	return NULL;
}


void find_rsdt(void)
{
	int fd;
	char *table_page, *table_ptr;
	struct rsdp_table *table;

	/* not yet initialized */
	if (!RSDP_ADDRESS)
		return;

	if (!e820_is_reserved(RSDP_ADDRESS)) {
		char buffer[1024];
		sprintf(buffer, "E820: RSDP (0x%lx) is not in reserved or ACPI memory!", (unsigned long)RSDP_ADDRESS);
		report_result(NULL, FAIL, buffer, NULL, "e820://");
		/* for RSDP the bios gap is special, linux won't use it anyway so continue to trust it */
		if (RSDP_ADDRESS<640*1024 || RSDP_ADDRESS>1024*1024)
			return;
	}

	fd = open("/dev/mem", O_RDONLY);
	if (fd<0)
		return;



	table_page = mmap(NULL, 2*getpagesize(), PROT_READ, MAP_SHARED, fd, RSDP_ADDRESS & ~(getpagesize()-1));

	if (!table_page)
		return;

	if (table_page==(void*)-1) {
		printf("Error is %s \n", strerror(errno));
		return;
	}


	table_ptr = table_page + (RSDP_ADDRESS & (getpagesize()-1));
	table = (struct rsdp_table *) table_ptr;


	if (strncmp(table->signature,"RSD PTR",7)!=0)
		return;

	rsdt = table->rsdt_address;
	xsdt = table->xsdt_address;
	rsdt_size = table->rsdt_length;

	munmap(table_page, 2*getpagesize());
	

	close(fd);
}

int locate_acpi_table_rsdt(char *name, unsigned long *address, unsigned long *size)
{
	char *table_ptr;
	struct generic_acpi_header *table;
	uint32_t *entries;
	int nr, i;

	*address = 0;
	*size = 0;

	if (!rsdt) 
		return -1;

	table_ptr = copy_acpi_table(rsdt, "RSDT");
	table = (struct generic_acpi_header *) table_ptr;

	if (table==NULL)
		return -1;

	nr = ((table->length - sizeof(struct generic_acpi_header))/4);

	if (nr<0) {
		report_result(NULL, FAIL, "RSDT table appears to have a negative number of entries", NULL, NULL);
		return -1;
	}

	entries = (uint32_t*)( (table_ptr + 36));

	for (i = 0; i<nr; i++) {
		if (!e820_is_reserved(entries[i])) {
			char buffer[1024];
			sprintf(buffer, "E820: RSDT table entry at 0x%lx is not in reserved/ACPI memory", (unsigned long)address);
			report_result(NULL, FAIL, buffer, NULL, "e820://");
			#ifndef OVERRIDE_E820
			return -1;
			#endif
		}
		table = copy_acpi_header(entries[i], name);
		if (table==NULL)
			return -1;

		if (strncmp(table->signature, name,4)==0) {
			*address = (unsigned long)entries[i];
			*size  = table->length;
		}
		free(table);
	}
	free(table_ptr);
	return 0;
}


int locate_acpi_table_xsdt(char *name, unsigned long *address, unsigned long *size)
{
	char *table_ptr;
	struct generic_acpi_header *table;
	uint64_t *entries;
	int nr, i;

	*address = 0;
	*size = 0;

	if (!xsdt) 
		return -1;

	table_ptr = copy_acpi_table(xsdt, "XSDT");
	table = (struct generic_acpi_header *) table_ptr;

	if (table==NULL)
		return -1;

	nr = ((table->length - sizeof(struct generic_acpi_header))/8);

	if (nr<0) {
		report_result(NULL, FAIL, "XSDT table appears to have a negative number of entries", NULL, NULL);
		return -1;
	}

	entries = (uint64_t*)( (table_ptr + 36));

	for (i = 0; i<nr; i++) {

		if (!e820_is_reserved(entries[i])) {
			char buffer[1024];
			sprintf(buffer, "E820: XSDT table entry at 0x%lx is not in reserved/ACPI memory", (unsigned long)address);
			report_result(NULL, FAIL, buffer, NULL, "e820://");
			#ifndef OVERRIDE_E820
			return -1;
			#endif
		}
		table = copy_acpi_header(entries[i], name);
		if (table==NULL)
			return -1;


		if (strncmp(table->signature, name,4)==0) {
			*address = (unsigned long)entries[i];
			*size  = table->length;
		}
		free(table);
	}

	free(table_ptr);
	return 0;
}




int locate_acpi_table(char *name, unsigned long *address, unsigned long *size)
{
	*address = 0;
	*size = 0;

	unsigned long x_addr = 0 , x_size = 0, r_addr = 0, r_size = 0;
	int ret_x = 0, ret_r = 0;

	/* not yet initialized */
	if (!rsdt && !xsdt) 
		find_rsdt();

	if (!rsdt && !xsdt) 
		return -1;
	
	if (xsdt)
		ret_x = locate_acpi_table_xsdt(name, &x_addr, &x_size);
	if (rsdt) 
		ret_r =  locate_acpi_table_rsdt(name, &r_addr, &r_size);

	if (xsdt && rsdt && ret_x != -1 && ret_r != -1) {
		if (x_addr != r_addr && strcmp(name,"FACP")!=0) {
			char buffer[1024];
			sprintf(buffer, "ACPI: there are different addresses (%lx and %lx) for table %s in XSDT versus RSDT.",
				x_addr, r_addr, name);
			report_result(NULL, WARN, buffer, NULL, NULL);
		}
	}
	if (xsdt && ret_x != -1) {
		*address = x_addr;
		*size = x_size;
		return ret_x;
	}
	*address = r_addr;
	*size = r_size;
	return ret_r;
}

int locate_acpi_table32(char *name, unsigned long *address, unsigned long *size)
{
	*address = 0;
	*size = 0;

	unsigned long r_addr = 0, r_size = 0;
	int ret_r = 0;

	/* not yet initialized */
	if (!rsdt) 
		find_rsdt();

	if (!rsdt) 
		return -1;
	
	if (rsdt) 
		ret_r =  locate_acpi_table_rsdt(name, &r_addr, &r_size);

	*address = r_addr;
	*size = r_size;
	return ret_r;
}


int locate_acpi_table64(char *name, unsigned long *address, unsigned long *size)
{
	*address = 0;
	*size = 0;

	unsigned long x_addr = 0 , x_size = 0;
	int ret_x = 0;

	/* not yet initialized */
	if (!xsdt) 
		find_rsdt();

	if (!xsdt) 
		return -1;
	
	ret_x = locate_acpi_table_xsdt(name, &x_addr, &x_size);

	*address = x_addr;
	*size = x_size;
	return ret_x;
}

#define DMESG_SIZE 128000

char *execute_aml_method(char *method, char *execute)
{
	FILE *file;
	char *dmesg;
	char *retval = NULL;

	unsigned long bytes_before;

	dmesg = malloc(DMESG_SIZE+1);
	if (!dmesg)
		return NULL;
	
	memset(dmesg, 0, DMESG_SIZE+1);
        klogctl (3, dmesg, DMESG_SIZE);
	bytes_before = strlen(dmesg);
	/*******************/

	file = fopen("/proc/acpi/hotkey/event_config", "w");
	if (!file) 
		goto out;
	fprintf(file, "%s", method);
	fclose(file);

	file = fopen("/proc/acpi/hotkey/action", "w");
	if (!file) 
		goto out;
	fprintf(file, "%s", execute);
	fclose(file);



	/*******************/
        klogctl (3, dmesg, DMESG_SIZE);
	retval = strdup(dmesg + bytes_before);

out:	
	free(dmesg);
	return retval;	
}


