/* 
 * Copyright (C) 2006, Intel Corporation
 * 
 * This file is part of the Linux-ready Firmware Developer Kit
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program in a file named COPYING; if not, write to the 
 * Free Software Foundation, Inc., 
 * 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include <sys/klog.h>

#include "biostest.h"


#define DMESG_SIZE 128000


GList *boot_dmesg;		/* Holds dmesg info, accessed by non-standalone tests */
extern char *current_test;


/* Obtains dmesg, cleans it up, and adds it to our resource list.
 * The boot_dmesg here is usually accessed by non-standalone tests
 * (i.e. *.so tests) */
void get_dmesg_buffer(void)
{
	char *dmesg;
	char *c1, *c2;
	int linecount = 0;
	dmesg = malloc(DMESG_SIZE);
	assert(dmesg!=NULL);

	system("/bin/cp -f " DMESG_LOCAL_FILE " " DMESG_FILE);

	/* don't printk to the console; newt hates that */
	klogctl (6, NULL, 0);

	memset(dmesg, 0, DMESG_SIZE);
	klogctl (4, dmesg, DMESG_SIZE);

	/* Add to our resource list that we keep. The user
	 * will have access to this when they look at the 
	 * test results and info,
	 * see uri.c for implementation */
	announce_resource("dmesg://", dmesg, NULL);

	c1 = dmesg;
	while (c1) {
		/* get next line from dmesg before '\n' 
		 * and put into 'c1' */
		c2 = strchr(c1, '\n');
		
		/* break if we're at the end of dmesg */
		if (c2==NULL)
			break;
		
		/* take out extra \n from beginning */
		*c2 = 0;
		c2++;
		
		/* skip the <4> that is in front of each line */
		if (strlen(c1)>3 && c1[0]=='<')
			c1+=3;

		/* now append the cleaned-up line to boot_dmesg */
		boot_dmesg = g_list_append(boot_dmesg, strdup(c1));
		linecount ++;
		
		/* put what's left of dmesg back into c1 */
		c1 = c2;
	}
	
	free(dmesg);

}

/* check for common dmesg errors */
void dmesg_common_errors(char *line, char *suggested_uri)
{
	char dmesg_uri[1024];

	if (current_test)
		sprintf(dmesg_uri, "dmesg://%s", current_test);
	else
		strcpy(dmesg_uri," dmesg://");

	if (suggested_uri)
		strcpy(dmesg_uri, suggested_uri);

	if (strstr(line, "Temperature above threshold, cpu clock throttled"))
		report_result(NULL, FAIL, "Test caused overtemperature. Insufficient cooling?", line, dmesg_uri);

	if (strstr(line, "ACPI: Handling Garbled _PRT entry"))
		report_result(NULL, FAIL, "Bios has a garbled _PRT entry; source_name and source_index swapped", line, dmesg_uri);

	if (strstr(line, "BIOS never enumerated boot CPU"))
		report_result(NULL, FAIL, "The boot processor is not enumerated!\n", line, dmesg_uri);

	if (strstr(line, "*** Error: Return object type is incorrect")) {
		char buf[8192];
		sprintf("Return object type is incorrect: %s \n", line);
		report_result(NULL, FAIL, buf, line, dmesg_uri);
	}
	if (strstr(line, "shpchp: acpi_shpchprm") && strstr(line,"_HPP fail"))
		report_result(NULL, FAIL, "Hotplug _HPP method fails!", line, dmesg_uri);
	if (strstr(line, "shpchp: acpi_pciehprm") && strstr(line,"OSHP fail"))
		report_result(NULL, FAIL, "Hotplug OSHP method fails!", line, dmesg_uri);
	if (strstr(line, "shpchp: acpi_shpchprm:") && strstr(line,"evaluate _BBN fail"))
		report_result(NULL, FAIL, "Hotplug _BBN method is missing", line, dmesg_uri);

	if (strstr(line, "[PBST] Namespace lookup failure, AE_NOT_FOUND")) 
		report_result(NULL, FAIL, "ACPI Namespace lookup failure reported", line, dmesg_uri);

	if (strstr(line, "*** Error: Method reached maximum reentrancy limit")) 
		report_result(NULL, FAIL, "ACPI method has reached reentrancy limit", line, dmesg_uri);

	if (strstr(line, "Error while parsing _PSD domain information"))
		report_result(NULL, FAIL, "_PSD domain information is corrupt!", line, dmesg_uri);

	if (strstr(line, "Wrong _BBN value, reboot and use option 'pci=noacpi'"))
		report_result(NULL, FAIL, "The BIOS has wrong _BBN value, "
			"which will make PCI root bridge have wrong bus number", line, dmesg_uri);
			
	if (strstr(line,"ACPI: lapic on CPU ") && strstr(line, " stops in C2[C2]"))
		report_result(NULL, FAIL, "The local apic timer incorrectly stops during C2 idle state",
				"The local apic timer incorrectly stops during C2 idle state. \n"
				"The ACPI specification forbids this and Linux needs the local \n"
				"apic timer to work. The most likely cause of this is that the \n"
				"firmware uses a hardware C3 or C4 state that is mapped to \n"
				"the ACPI C2 state.", dmesg_uri);

	if (strstr(line, "Invalid _PCT data")!=NULL) {
		char uri[4096];
		char buf[4096];
		char line2[4096];
		char *c, *c2;
		memset(uri, 0, 4096);
		memset(buf, 0, 4096);
		memset(line2, 0, 4096);
		strcpy(line2,line);		
		c = strchr(line2, '[');
		if (c)
			c = strchr(c+1, '[');
		if (c) {
			c++;
			c2=strchr(c,']');
			if (c2) *c2=0;
			AML_to_uri(c, uri);
			sprintf(buf,"The _PCT data of %s is invalid", c);
		} else sprintf(buf, "The _PCT data is invalid");
		report_result(NULL, FAIL, buf, line, uri);
	}

	if (strstr(line, "*** Error: Method execution failed")) {
		char uri[4096];
		char buf[4096];
		char line2[4096];
		char *c, *c2;
		memset(uri, 0, 4096);
		memset(buf, 0, 4096);
		memset(line2, 0, 4096);
		strcpy(line2,line);		
		c = strchr(line2, '[');
		if (c) {
			c++;
			c2=strchr(c,']');
			if (c2) *c2=0;
			
			AML_to_uri(c, uri);
			sprintf(buf,"Method execution of %s failed", c);
		} else strcpy(buf, line);
		report_result(NULL, FAIL, buf, line, uri);
	}

	if (strstr(line, "Method parse/execution failed") && strstr(line, "AE_NOT_FOUND")) {
		char uri[4096];
		char buf[4096];
		char line2[4096];
		char *c, *c2;
		memset(uri, 0, 4096);
		memset(buf, 0, 4096);
		memset(line2, 0, 4096);
		strcpy(line2,line);		
		c = strchr(line2, '[');
		if (c) {
			c++;
			c2=strchr(c,']');
			if (c2) *c2=0;
			
			AML_to_uri(c, uri);
			sprintf(buf,"Method parsing/execution of %s failed", c);
		} else strcpy(buf, line);
		report_result(NULL, FAIL, buf, line, uri);
	}
	if (strstr(line, "*** Error: Method execution failed") && strstr(line, "AE_AML_METHOD_LIMIT")) {
		char uri[4096];
		char buf[4096];
		char line2[4096];
		char *c, *c2;
		memset(uri, 0, 4096);
		memset(buf, 0, 4096);
		memset(line2, 0, 4096);
		strcpy(line2,line);		
		c = strchr(line2, '[');
		if (c) {
			c++;
			c2=strchr(c,']');
			if (c2) *c2=0;
			
			AML_to_uri(c,uri);
			sprintf(buf,"Method %s reached maximum reentrancy limit", c);
		} else strcpy(buf, line);
		report_result(NULL, FAIL, buf, line, uri);
	}
		
	if (strstr(line, "Method execution failed") && strstr(line, "AE_OWNER_ID_LIMIT")) {
		char uri[4096];
		char buf[4096];
		char line2[4096];
		char *c, *c2;
		memset(uri, 0, 4096);
		memset(buf, 0, 4096);
		memset(line2, 0, 4096);
		strcpy(line2,line);		
		c = strchr(line2, '[');
		if (c) {
			c++;
			c2=strchr(c,']');
			if (c2) *c2=0;
			
			AML_to_uri(c, uri);
			sprintf(buf,"Method %s failed to allocate owner ID", c);
		} else strcpy(buf, line);
		report_result(NULL, FAIL, buf, line, uri);
	}

	if (strstr(line, "Method execution failed") && strstr(line, "AE_AML_BUFFER_LIMIT")) {
		char uri[4096];
		char buf[4096];
		char line2[4096];
		char *c, *c2;
		memset(uri, 0, 4096);
		memset(buf, 0, 4096);
		memset(line2, 0, 4096);
		strcpy(line2,line);		
		c = strchr(line2, '[');
		if (c) {
			c++;
			c2=strchr(c,']');
			if (c2) *c2=0;
			
			AML_to_uri(c, uri);
			sprintf(buf,"Method %s failed: ResourceSourceIndex is present but ResourceSource is not", c);
		} else strcpy(buf, line);
		report_result(NULL, FAIL, buf, line, uri);
	}

	if (strstr(line, "Disabling IRQ")!=NULL) 
		report_result(NULL, FAIL,"The kernel detected an irq storm. IRQ routing bug?", line, dmesg_uri);
	
		
}

/* Checks the kernel msg buffer for any new messages.  Usually called 
 * after a test finished (finish_test()) to see if it resulted in
 * any msgs. */
void check_dmesg_updates(char *uri)
{
	char *dmesg;
	char *c1, *c2;
	int linecount = 0;
	char buf[1024];
	

	dmesg = malloc(DMESG_SIZE);
	assert(dmesg!=NULL);

	memset(dmesg, 0, DMESG_SIZE);
	klogctl (4, dmesg, DMESG_SIZE);

	if (current_test)
		sprintf(buf, "dmesg://%s", current_test);
	else
		strcat(buf, "dmesg://");

	announce_resource(buf, dmesg, "dmesg://");

	c1 = dmesg;
	while (c1) {
		c2 = strchr(c1, '\n');
		if (c2==NULL)
			break;
		*c2 = 0;
		c2++;
		/* skip the <x> that is in front of each line
		 * if it exists. */
		if (strlen(c1)>3 && c1[0]=='<')
			c1+=3;

		dmesg_common_errors(c1, uri);
		linecount ++;

		c1 = c2;
	}
	free(dmesg);

}
