/* 
 * Copyright (C) 2006, Intel Corporation
 * 
 * This file is part of the Linux-ready Firmware Developer Kit
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation;version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program in a file named COPYING; if not, write to the 
 * Free Software Foundation, Inc., 
 * 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>

#include <glib.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <firmwarekit.h>


struct dsdt_line {
	char *line;		/* actual line in dsdt */
	char *locator;		/* parsed and cleaned locator (e.g. \GPE._L01) */
	char *untranslated; 	/* non-parsed locator (e.g. @\_GPE{_L01) */
	int lineno;		/* line # in dsdt */
};

#define SCOPE 1
#define DEVICE 2
#define METHOD 3

GList *ssdt[MAX_SSDTS];		/* holds the list of available dsdt and ssdt* tables */
int ssdt_size[MAX_SSDTS] = {4,4,4,4,4,4,4,4,4,4,4};
static int current_ssdt;	/* current pointer to which dsdt or ssdt table we're accessing */

/* extract_dsdt_ssdts () -- Obtain the dsdt and ssdt acpi tables */
void extract_dsdt_ssdts()
{

	int ret;
	
	/* create hex-dump format of all acpi tables in file 'acpidump' */
	system("/usr/sbin/acpidump > /var/log/firmwarekit/acpi.dump &> /dev/null");
	if (access("/var/log/firmwarekit/acpi.dump", R_OK))
		fprintf(stderr,"WARN (acpidump): failed to create /var/log/firmwarekit/acpi.dump.\n");
	
	/* use 'acpidump' file to extract dsdt and ssdt tables
	 * in binary format, creates DSDT.dat and SSDT*.dat */
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/sbin/acpixtract acpi.dump &> /dev/null; popd >/dev/null");
	if (access("/var/log/firmwarekit/DSDT.dat", R_OK)) {
		ret = system("cat /proc/acpi/dsdt > /var/log/firmwarekit/DSDT.dat");
		if (ret != EXIT_SUCCESS) {
			fprintf(stderr,"WARN (acpixtract: failed to create DSDT.dat.\n");
			return;
		}
	}

	/* Disassemble DSDT.dat with iasl, will create DSDT.dsl */
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d DSDT.dat &>/dev/null; popd >/dev/null");
	if (access("/var/log/firmwarekit/DSDT.dsl", R_OK))
		fprintf(stderr,"WARN (iasl): failed to create DSDT.dsl.\n");

	/* Include ACPI table for external symbol resolution (if they exist) */
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d -e DSDT.dat SSDT.dat &>/dev/null; popd >/dev/null");
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d -e DSDT.dat SSDT1.dat &>/dev/null; popd >/dev/null");
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d -e DSDT.dat SSDT2.dat &>/dev/null; popd >/dev/null");
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d -e DSDT.dat SSDT3.dat &>/dev/null; popd >/dev/null");
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d -e DSDT.dat SSDT4.dat &>/dev/null; popd >/dev/null");
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d -e DSDT.dat SSDT5.dat &>/dev/null; popd >/dev/null");
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d -e DSDT.dat SSDT6.dat &>/dev/null; popd >/dev/null");
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d -e DSDT.dat SSDT7.dat &>/dev/null; popd >/dev/null");
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d -e DSDT.dat SSDT8.dat &>/dev/null; popd >/dev/null");
	system("pushd /var/log/firmwarekit/ >/dev/null; /usr/bin/iasl -d -e DSDT.dat SSDT9.dat &>/dev/null; popd >/dev/null");

	/* current iasl has an issue with arugment counts in SSDT's, work around this */
	system("sed -i -e \"s/^ACPI Error.*Argument count mismatch for method.*//g\" /var/log/firmwarekit/SSDT*.dsl &> /dev/null");

}

/* push () -- adds a '{' to locator string */
static void push(char *s)
{
	if (strlen(s)>8000)  {
		printf("FATAL: locator overflow\n");
		return;
	}
	strcat(s, "{");
}

/* push_name() -- adds label name to locator string */
static void push_name(char *s, char *name)
{
	if (strlen(s)+strlen(name)>8000) {
		printf("FATAL: locator overflow\n");
		return;
	}
	strcat(s,name);
}

/* push_name_scope() -- adds '@' . 'scope name'
 *  to locator string (using @ to diff between scope
 *  and label like Method, Device, etc.) */
static void push_name_scope(char *s, char *name)
{
	if (strlen(s)+strlen(name)>8000) {
		printf("FATAL: locator overflow\n");
		return;
	}
	strcat(s,"@");
	strcat(s,name);
}

/* pop () - chops chars until we reach a '{' */
static void pop(char *s)
{
	if (strlen(s)<1)
		return;
	if (s[strlen(s)-1]!='{')
		return;
	chop_lastchar(s);
	while (strlen(s)>0 && s[strlen(s)-1]!='{')
		chop_lastchar(s);
}

/* translate_locator() -- parses the locator and cleans it up to 
 * produce the actual AML path. (e.g. @\GPE{_L01 becomes
 * \GPE.L01) */
static void translate_locator(char *locator, char *target) 
{
	char *c;
	char *l;
	int prevbracket = 1;

	if (target == NULL)
		return;

	target[0] = 0;

	if (locator == NULL)
		return;

	c = target;

	l=locator;

	while (*l) {
		if (*l=='{' && !prevbracket) {
			*c = '.';
			c++;
			*c = 0;
			prevbracket = 1;
		}
		if (*l!='{') {
			*c = *l;
			c++;
			*c = 0;
			prevbracket = 0;
		}
		if (*l=='@') {
			c = target;
			*c=0;
			prevbracket = 1;
		}
	
		l++;
	}
}

/* parse_dsdt () -- parses dsdt.dsl file.  Saves the info of each line as
 * 			a list of struct* dsdt_line.
 * 			This makes it easy later on when we recompile the 
 * 			DSDT with iasl to identify which line the 
 * 			warnings/errors are, and which methods/
 * 			devices/scope they're located in. 
 * 							
 * 
 * Parameters
 *   	filename - usually dsdt.dsl file to parse
 * Returns
 * 		void
*/

static void parse_dsdt(char *filename)
{
	FILE *file;
	char line[4096];
	int lineno = 0;

	char locator[8192];
	char translated[8192];

	memset(locator, 0, 8192);

	/* open dsdt.dsl or ssdt*.dsl */
	file = fopen(filename, "r");
	if (file==NULL) {
		printf("Failed to open file %s \n", filename);
		return;
	}	
	
  	/* Parse through the .dsl file */
	while (!feof(file)) {
		char *c;
		char name[256];
		struct dsdt_line *ds;
		int label = 0;

		memset(name, 0, 256);
		memset(line, 0, 4096);
		if (fgets(line, 4095, file)==NULL)
			break;
		lineno++;
		
		ds = malloc(sizeof(struct dsdt_line));
		assert(ds!=NULL);		
		memset(ds, 0, sizeof(struct dsdt_line));
		ds->line = strdup(line);
		ds->lineno = lineno;

		/* We hit a Device label */
		c = strstr(line, "Device (");
		if (c) {

			/* get the name of the device */
			label = DEVICE;
			c+=8;
			while (*c) {
				if (*c!=')' && *c!=',') 
					name[strlen(name)]=*c;
				else 
					break;
				c++;
			}
			push_name(&locator[0], name);
		}

		/* We hit a Method label */
		c = strstr(line, "Method (");
		if (c) {

			label = METHOD;
			c+=8;
			while (*c) {
				if (*c!=')' && *c!=',') 
					name[strlen(name)]=*c;
				else 
					break;
				c++;
			}
			push_name(&locator[0], name);
		}
		
		/* We hit a Scope label */
		c = strstr(line, "Scope (");
		if (c) {

			label = SCOPE;
			c+=7;
			while (*c) {
				if (*c!=')' && *c!=',') 
					name[strlen(name)]=*c;
				else 
					break;
				c++;
			}
			push_name_scope(&locator[0], name);
		}

		/* Else if we didn't hit a label */
		if (label == 0) {
			c = line;
			while (c && *c) {
				if (*c=='{')
					push(&locator[0]);
				if (*c=='}')
					pop(&locator[0]);
				c++;
			}
		}
		
		translate_locator(locator, translated); 
		ds->locator = strdup(translated);
		ds->untranslated = strdup(locator);
		ssdt[current_ssdt] = g_list_append(ssdt[current_ssdt], ds);
		ssdt_size[current_ssdt] += strlen(ds->line);
	}
	fclose(file);
}


/* do_new_resource() -- goes through our list of available dsdt and ssdt* 
 * 			tables, and calls the function to announce the resources 
 *			within them.
 * Parameters
 *   	list - pointer to the current dsdt or ssdt* table held by
 * Returns
 * 		void
*/
static void do_new_resource(GList *lst)
{	
	GList *list;
	struct dsdt_line *line;
	char *buffer = NULL;
	char *ptr;
	int state = 0;
	char uri[2048];
	char parent[2048];
	int skip = 100;

	if (!lst)
		return;

	line = (struct dsdt_line*)lst->data;

	ptr =buffer = malloc(ssdt_size[current_ssdt]);

	ptr += sprintf(ptr,"--=[ %s ]=--\n\n", line->locator);

	list = lst;
	while (list) {
		struct dsdt_line *this;

		this = (struct dsdt_line*) list->data;

		if (strlen(line->locator)>3) {
			int Q;
			Q = space_count(this->line);
			if (Q<skip)
				skip = Q;
			ptr += sprintf(ptr, "%s", &this->line[skip]);
		}

		if (!is_prefix(line->untranslated, this->untranslated))
			break;

		if (state==0 && strcmp(line->untranslated, this->untranslated)!=0) {
			state=1;
			do_new_resource(list);
		}
		if (strcmp(line->untranslated, this->untranslated)==0)
			state = 0;			
		list = g_list_next(list);
	}

	AML_to_uri(line->locator, uri);

	strcpy(parent,uri);
	while (strlen(parent)>0 && parent[strlen(parent)-1]!='.')
			parent[strlen(parent)-1] = 0;

	chop_lastchar(parent);

	if (strlen(line->locator)>=2 && line->locator[strlen(line->locator)-1]!='.')
		announce_resource(uri, buffer, parent);

	free(buffer);
}

/* announce_aml_resources() -- goes through our list of available dsdt and ssdt* 
 * 			tables, and calls the function to announce the resources 
 * 			within them.  */
static void announce_aml_resources(void)
{
	if (ssdt[current_ssdt]) 
		do_new_resource(ssdt[current_ssdt]);
	current_ssdt++;
}


/* get_dsdt_resources() -- obtains the dsdt and ssdt* tables, parses the resources
 * 			within, announces the resources, and saves the information.
 * 
 * Parameters
 *   	void
 * Returns
 * 	void
*/

void get_dsdt_resources(void)
{
	int i;
	char filen[1024];

	extract_dsdt_ssdts();
	  
	parse_dsdt("/var/log/firmwarekit/DSDT.dsl");
	announce_aml_resources();

	for (i=1; i<10; i++) {
		sprintf(filen, "/var/log/firmwarekit/SSDT%i.dsl", i);
		if (!access(filen, R_OK)) {
			parse_dsdt(filen);
			announce_aml_resources();
		}
	}


}
