/* 
 * Copyright (C) 2006, Intel Corporation
 * 
 * This file is part of the Linux-ready Firmware Developer Kit
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program in a file named COPYING; if not, write to the 
 * Free Software Foundation, Inc., 
 * 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>


#include <glib.h>

#include "biostest.h"


struct e820_entry {
	uint64_t	start_address;
	uint64_t	end_address;
	int		type;
};

static GList *e820;

static int have_e820;

void register_e820_line(uint64_t start, uint64_t end, int type)
{
	struct e820_entry *entry;
	entry = malloc(sizeof(struct e820_entry));
	assert(entry!=NULL);
	memset(entry, 0, sizeof(struct e820_entry));

	entry->start_address = start;
	entry->end_address = end;
	entry->type = type;

	e820 = g_list_append(e820, entry);
	have_e820 = 1;
}


int e820_type(uint64_t memory)
{
	int result = E820_UNKNOWN;
	GList *list;
	struct e820_entry *entry;

	list = g_list_first(e820);
	while (list) {
		entry = (struct e820_entry*)list->data;

		if (entry->start_address <= memory && entry->end_address > memory) 
			result = entry->type;

		list = g_list_next(list);
	}

	return result;
}

int e820_is_reserved(uint64_t memory)
{
	int result = E820_UNKNOWN;
	GList *list;
	struct e820_entry *entry;
	
	/* when we don't have E820 info, assume all is fair */
	if (!have_e820)
		return 1; 

	/* bios data area is always reserved */
	if (memory >= 640 * 1024 && memory <= 1024*1024)
		return 1;

	list = g_list_first(e820);
	while (list) {
		entry = (struct e820_entry*)list->data;

		if (entry->start_address <= memory && entry->end_address > memory) 
			result = entry->type;

		list = g_list_next(list);
	}

	if (result == E820_RESERVED)
		return 1;
	if (result == E820_ACPI)
		return 1;
	return 0;
}

/* checks dmesg for useful e820 info */
static void e820_dmesg_info (gpointer data, gpointer user_data)
{
	char *line = (char *)data;

	if (strstr(line, "ACPI") && strstr(line,"RSDP")) {
		char *c;
		c = strstr(line, "@ 0x");
		if (c) 
			RSDP_ADDRESS = strtoul(c+4, NULL, 16);
	}

	if (strstr(line,"BIOS-e820:")) {
		uint64_t start, end;
		char *c;
		int type = E820_UNKNOWN;
		start = strtoull(&line[13], NULL, 16);
		c= strstr(line," - ");
		if (c) c += 3;
		end   = strtoull(c, NULL, 16);
		if (strstr(line, "(usable)"))
			type = E820_USABLE;
		if (strstr(line, "(reserved)"))
			type = E820_RESERVED;
		if (strstr(line, "ACPI"))
			type = E820_ACPI;

		register_e820_line(start, end, type);
	}
}

void e820_register_resource(void)
{
	/* First, check dmesg for useful e820 info */
	g_list_foreach(boot_dmesg, e820_dmesg_info, NULL);

	char *buffer = strdup("E820 memory layout table\n-------------------------------------\n");
	GList *list;
	struct e820_entry *entry;
	list = g_list_first(e820);
	while (list) {
		entry = (struct e820_entry*)list->data;
		buffer = scatprintf(buffer,"%016llx - %016llx    ", entry->start_address, entry->end_address);

		if (entry->type == E820_RESERVED)
			buffer = scatprintf(buffer,"<reserved>");
		if (entry->type == E820_ACPI)
			buffer = scatprintf(buffer,"<ACPI use>");
		if (entry->type == E820_USABLE)
			buffer = scatprintf(buffer,"OS used ram");
			
		buffer = scatprintf(buffer,"\n");

		list = g_list_next(list);
	}
	announce_resource("e820://", buffer, "dmesg://");

	free(buffer);
}

