/* 
 * Copyright (C) 2006, Intel Corporation
 * 
 * This file is part of the Linux-ready Firmware Developer Kit
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation;version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program in a file named COPYING; if not, write to the 
 * Free Software Foundation, Inc., 
 * 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include <sys/klog.h>

#include "biostest.h"


GList *boot_dmesg = NULL;	/* Holds dmesg info, accessed by standalone tests */	
GList *ssdt[MAX_SSDTS];		/* Holds DSDT and SSDT tables, accessed by stndaln tests */


/* encode the output of the test results from ascii to hex */
static char *encode(char *string)
{
	char *buf;
	char *hex="0123456789ABCDEF";
	int len;
	int count = 0;


	if (string==NULL)
		return strdup("--");

	if (getenv("DEBUG")!=NULL) {
		return strdup(string);
	}

	len = strlen(string);
	buf = malloc(len*2+1);

	while (*string) {
		buf[count++] = hex[((unsigned char)*string) >> 4];
		buf[count++] = hex[((unsigned char)*string) & 15];
		string++;
	}
	buf[count]=0;
	return buf;
}


void start_test(char *testID, char *name, char *description) 
{
	char *test_id = encode(testID);
	char *_name = encode(name);
	char *_description = encode(description);
	fprintf(stdout, "S %s %s %s\n", test_id, _name, _description);
	free(test_id);
	free(_name);
	free(_description);
	fflush(stdout);
}

void report_result(char *testID, int res, char *summary, char *details, char *devuri)
{
	char result[16];
	char *test_id = encode(testID);
	char *_result;
	char *_summary = encode(summary);
	char *_details = encode(details);
	char *_devuri = encode(devuri);


	memset(result, 0, 16);

	switch (res) {
	case PASS:	
		strcpy(result, "pass");
		break;
	case WARN:
		strcpy(result, "warn");
		break;
	case INFO:
		strcpy(result, "info");
		break;
	case FAIL:
		strcpy(result, "fail");
		break;
	default:
		strcpy(result, "BUG!");
		break;
	}
	_result = encode(result);

	fprintf(stdout, "R %s %s %s %s %s\n", test_id, _result, _summary, _details, _devuri);
	free(test_id);
	free(_result);
	free(_summary);
	free(_details);
	free(_devuri);
	fflush(stdout);
}

void finish_test(char *testID)
{
	char *test_id = encode(testID);
	fprintf(stdout, "F %s\n", encode(testID));
	free(test_id);
	fflush(stdout);
}

void report_testrun_progress(int percentage)
{
	fprintf(stdout, "P %i\n", percentage);
	fflush(stdout);
}

void announce_resource(char *uri, char *description, char *parent)
{
	char *_uri = encode(uri);
	char *desc = encode(description);
	char *par = encode(parent);

	fprintf(stdout, "U %s %s %s\n", _uri, desc, par);
	free(_uri);
	free(desc);
	free(par);

	fflush(stdout);
}

/* Obtains the dmesg kernel buffer. The boot_dmesg here is usually
 * accessed by standalone tests (i.e. *.exe) */
void load_boot_dmesg_buffer(void)
{
	char line [4096];
	FILE *file;

	if(boot_dmesg != NULL)
		return;

	memset(line, 0, 4096);

	/* First, try to access the tmp dmesg output that
	 * get_dmesg_buffer() created. If it doesn't exist,
	 * we're in standlone mode, and we should access
	 * the local dmesg file. */	
	if (access(DMESG_FILE, R_OK))
		file = fopen(DMESG_LOCAL_FILE, "r");
	else
		file = fopen(DMESG_FILE, "r");

	if (!file)
		return;

	while (!feof(file)) {
		if (fgets(line, 4095, file)==NULL)
			break;
		boot_dmesg = g_list_append(boot_dmesg, strdup(line));
	}

	fclose(file);
}

void load_dsdt_ssdts(void) {

	int ret, i;
	static int current_ssdt = 0;
	char command[4096];
	char line[4096];
	char filen[1024];
	FILE *file;


	/* create hex-dump format of all acpi tables in file 'acpi.dump' */
	sprintf(command, "%sacpidump > /var/log/firmwarekit/acpi.dump &> /dev/null", "/usr/sbin/");
	system(command);
	if (access("/var/log/firmwarekit/acpi.dump", R_OK))
		fprintf(stderr,"WARN (acpidump): failed to create acpi.dump.\n");
	
	/* use 'acpidump' file to extract dsdt and ssdt tables
	 * in binary format, creates DSDT.dat and SSDT*.dat */
	sprintf(command, "pushd /var/log/firmwarekit/ >/dev/null; %sacpixtract acpi.dump &> /dev/null; popd >/dev/null", "/usr/sbin/");
	system(command);
	if (access("/var/log/firmwarekit/DSDT.dat", R_OK)) {
		ret = system("cat /proc/acpi/dsdt > /var/log/firmwarekit/DSDT.dat");
		if (ret != EXIT_SUCCESS) {
			fprintf(stderr,"WARN (acpixtract): failed to create DSDT.dat.\n");
			return;
		}
	}

	/* Disassemble DSDT.dat with iasl, will create DSDT.dsl */
	sprintf(command, "pushd /var/log/firmwarekit/ >/dev/null; %siasl -d DSDT.dat &>/dev/null; popd >/dev/null", "/usr/bin/");
	system(command);
	file = fopen("/var/log/firmwarekit/DSDT.dsl", "r");
	if (!file)
		fprintf(stderr,"WARN (iasl): failed to create DSDT.dsl.\n");

	/* Save the contents of the dsdt in our list */
	while (!feof(file)) {
		if (fgets(line, 4095, file)==NULL)
			break;
		ssdt[current_ssdt] = g_list_append(ssdt[current_ssdt], strdup(line));
	}
	fclose(file);
	current_ssdt++;
					

	 /* Now extract the acpi tables for external symbol resolution 
	  * (if they exist, should use SSDT{1-9}.dat to create SSDT{1-9}.dsl) */
	 for(i=0;i<MAX_SSDTS;i++) {
		 if(i == 0)
			 sprintf(filen, "SSDT");
		 else
			 sprintf(filen, "SSDT%d", i);
			 
	 	sprintf(command, "pushd /var/log/firmwarekit/ >/dev/null; %siasl -d -e DSDT.dat %s.dat &>/dev/null; popd >/dev/null", "/usr/bin/", filen);
	 	system(command);

		strcat(filen, ".dsl");
		file = fopen(filen, "r");
		if (!file)
			break;

		/* Save the contents of the ssdt in our list */
		while (!feof(file)) {
			if (fgets(line, 4095, file)==NULL)
				break;
			ssdt[current_ssdt] = g_list_append(ssdt[current_ssdt], strdup(line));
		}
		fclose(file);
		current_ssdt++;
	 }
	 
	/* current iasl has an issue with arugment counts in SSDT's, work around this */
	system("sed -i -e \"s/^ACPI Error.*Argument count mismatch for method.*//g\" /var/log/firmwarekit/SSDT*.dsl &> /dev/null");

}

