/* 
 * Copyright (C) 2006, Intel Corporation
 * 
 * This file is part of the Linux-ready Firmware Developer Kit
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation;version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program in a file named COPYING; if not, write to the 
 * Free Software Foundation, Inc., 
 * 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>


#include <slang.h>
#include <newt.h>

#include <glib.h>

#include "biostest.h"


int use_serial;
static unsigned long serial_speed;
static int tty;

int safe_mode;
int interactive_mode = 1;
int saveovernet = 1;
extern char ssh_dest[];

int finished_tests = 0;

/* Parse the kernel parameter we got from func() parse_kernel_cmdline. 
 * It looks for console=ttySx, safe mode and noninteractive mode. */
static void parse_kernel_param(const char *buffer)
{
	const char *tmp;
	int ret;

	/* parse console=ttyS boot param */
	if (strstr(buffer, "console=ttyS")){
		tmp = buffer + 12;
		tty = (*tmp) - 48;
		if (tty < 0 || tty > 9)
			goto bad_param;
		if (strlen(buffer) > 14){
			ret = sscanf (buffer, "console=ttyS%d,%lu",
				      &tty, &serial_speed);
			if (ret < 2)
				goto bad_param;
			if (serial_speed != 9600 &&
			    serial_speed != 19200 &&
			    serial_speed != 38400 &&
			    serial_speed != 57600 &&
			    serial_speed != 115200){
				fprintf (stderr, "Invalid speed: %lu",
					 serial_speed);
				goto bad_param;
			}
		}
		use_serial=1;
	}

	/* parse safemode boot param */
	if (strstr(buffer, "safemode"))
		safe_mode=1;
	/* parse noninteractive boot param, default is 1*/
	if (strstr(buffer, "noninteractive"))
		interactive_mode=0;
	if (strstr(buffer, "saveovernet")) {
		saveovernet=1;
		ret = sscanf (buffer, "saveovernet=%s",
			      ssh_dest);
		if(ret < 1)
			goto bad_param;
        }
	return;

 bad_param:
	fprintf(stderr, "Illegal console parameter: %s",
		buffer);
	use_serial=0;
}

/* Goes through the kernel command options found in
 * /proc/cmdline', parses out a command, then calls the 
 * func() parse_kernel_param to parse this command. */
static void parse_kernel_cmdline(void)
{
	FILE *file;
	char buffer[4096];
	char param[4096];
	int x,last_param;
	const char *tmp;

	memset(buffer, 0, 4096);
	memset(param, 0, 4096);

	/* checking for kernel command line options in 
	   '/proc/cmdline' and put it in 'buffer' */

	file = fopen("/proc/cmdline","r");
	if (!file)
		return;
	if (fgets(buffer, 4095, file)==0)
		return;
	fclose(file);

	last_param = 0;

	/* copy buffer with cmd line options to 'tmp' */
	tmp = buffer;

	/* going thru all tmp array (which has the 'buffer' 
	   contents) */ 
	for (x = 0; x < 4096; x++){
		if (*(tmp + x) == '\t' || *(tmp + x) == ' ' || *(tmp + x) == '\n'
		    || *(tmp + x) == '\0'){
			
			/* 'param' will now have the parsed parameter */
			memcpy(param, (tmp + last_param), x - last_param);
			
			/* put '\0' to end the param array */
			param[x - last_param] = '\0';
			
			/*
			fprintf(stderr, "Found param from offset: %d to %d",
				last_param, x);
			fprintf(stderr, " length: %d\nParam: %s\n",
				x - last_param, param);
			*/

			/* now parse this 'param' itself */
			parse_kernel_param(param);
		}

		/* move last_param array position holder until we hit
		 * another parameter */
		while (*(tmp + x) == '\t' || *(tmp + x) == ' '){
			x++;
			last_param = x;
		}

		/* no more kernel params to parse */
		if (*(tmp + x) == '\n'
		    || (tmp + x) == '\0')
			break;
	}
}

int main(int argc, char **argv)
{
	/* initialisation of core support routines */

	parse_kernel_cmdline();

	/* if 'serial' arg was passed to main, then use
	 * serial.  This can also be set by func()
	 * parse_kernel_param if it's in the kernel
	 * cmdline options */
	if (argc>1 && strstr(argv[1],"serial"))
		use_serial = 1;

	/* initialize serial output (if enabled) 
	 * (implemented in serial.c) */
	init_serial(use_serial, tty, serial_speed);

	/* get and parse demsg for useful info
	 * (implemented in dmesg.c) */
	get_dmesg_buffer();
	
	/* obtain dsdt and ssdt acpi tables' resources */
	get_dsdt_resources();

	/* register e820 resource */
	e820_register_resource();

	/* display welcome msg, init newt root window 
	 * (implemented in ui.c) */
	init_results_ui();	

	/* execute all tests in 'plugins' dir 
	 * (implemented in plugins.c) */
	execute_plugins(LIBDIR "/firmwarekit/plugins/");

	finished_tests = 1;
	if(!interactive_mode) {
	  	dump_xml_results();
		if (ssh_dest[0] != 'h'){
	    		int x;
	    		x = copy_via_ssh(ssh_dest);
	    		if (x){
		      		fprintf(stderr, "ERROR: scp copy to %s did"
		      		" not succeed\n", ssh_dest);
	      			return EXIT_FAILURE;
	    		}
	    		else{
	      			fprintf(stdout, "Copying data to %s "
		      		"successfull\n", ssh_dest);
	    		}
	    	/* try to exit graphic stuff gracefully */
	  	}
	  	newtRefresh();
	  	newtPopWindow();
	  	newtFinished();
        }
        else{
		dump_results();
	}

	return EXIT_SUCCESS;
}
