/* 
 * Copyright (C) 2006, Intel Corporation
 * 
 * This file is part of the Linux-ready Firmware Developer Kit
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation;version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program in a file named COPYING; if not, write to the 
 * Free Software Foundation, Inc., 
 * 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <dlfcn.h>

#include <sys/types.h>
#include <dirent.h>


#include <glib.h>

#include "biostest.h"

typedef void   (plugin_func) (void);

/* perform 'run_test' function in the given .so file */
static void do_plugin(char *filename)
{
	void *handle;
	plugin_func *function;

	handle = dlopen(filename, RTLD_NOW);

	if (!handle) {
		char buf[4095];
		sprintf(buf, "Invalid plugin: %s :%s \n", filename, dlerror());
		non_fatal_bug(buf);
		return;	
	}

	function = dlsym(handle, "run_test");
	if (!function) {
		char buf[4095];
		sprintf(buf, "Invalid plugin: %s: No run_test function \n", filename);
		non_fatal_bug(buf);
		return;
	}
	function();
}

/* decode the output of running a plugin test from hexadecimal to ascii. */
static char *decode(char *string)
{
	char *buffer;
	int count = 0;

	if (strcmp(string,"--")==0)
		return NULL;

	buffer = malloc(strlen(string)+1);
	assert(buffer != NULL);
	while (*string) {
		int value = 0;
		int c = *string;
		if ((c>='0') && (c<='9'))
			value = c-'0';
		if ((c>='A') && (c<='F'))
			value = 10+(c-'A');
		if ((c>='a') && (c<='f'))
			value = 10+(c-'a');
		string++;
		if (! *string)
			break;
		c = *string;
		value = value << 4;
		if ((c>='0') && (c<='9'))
			value |= c-'0';
		if ((c>='A') && (c<='F'))
			value |= 10+c-'A';
		if ((c>='a') && (c<='f'))
			value |= 10+c-'a';
		buffer[count++] = value;
		string++;
	}
	buffer[count++] = 0;
	return buffer;
}

/* execute the test plugin (usually a .exe, .pl 
 * or .sh file) */
static void do_exe_plugin(char *filename)
{
	FILE *file;
	char buffer[4096];
	char *line;
	int ret;

	/* execute the file and return a pointer
	 * to the output stream */
	file = popen(filename, "r");
	if (!file)
		return;

	/* parse the output of the test */
	while (!feof(file)) {
		line = NULL;
repeat:
		memset(buffer, 0, 4096);
		fgets(buffer, 4096, file);
		line = scatprintf(line, "%s", buffer);
		if (strlen(buffer)==4095 && buffer[4095]!='\n')
			goto repeat;
		if (strlen(line)<2)
			continue;

		/* we've got a line that shows a result */
		if (line[0] == 'R') {
			char *testID, *res, *summary, *details, *devuri;
			int result = 0;
			testID = &line[2];

			res = strchr(&line[2], ' ');
			if (!res)
				continue;
			*res=0;
			res++;

			summary = strchr(res, ' ');
			if (!summary)
				continue;
			*summary=0;
			summary++;


			details = strchr(summary, ' ');
			if (!details)
				continue;
			*details=0;
			details++;


			devuri = strchr(details, ' ');
			if (!devuri)
				continue;
			*devuri=0;
			devuri++;

			res = decode(res);
		
			if (strstr(res, "pass"))
				result = PASS;
			if (strstr(res, "fail"))
				result = FAIL;
			if (strstr(res, "warn"))
				result = WARN;
			if (strstr(res, "info"))
				result = INFO;
			report_result(decode(testID), result, decode(summary), decode(details), decode(devuri));
		}

		/* Start of test */
		if (line[0] == 'S') {
			char *testID, *name, *description;
			testID = &line[2];

			name = strchr(&line[2], ' ');
			if (!name)
				continue;
			*name=0;
			name++;

			description = strchr(name, ' ');
			if (!description)
				continue;
			*description=0;
			description++;

			start_test(decode(testID), decode(name), decode(description));
		}

		/* the test is announcing a resource in our resourse list */
		if (line[0] == 'U') {
			char *uri, *description, *parent;
			uri = &line[2];

			description = strchr(&line[2], ' ');
			if (!description)
				continue;
			*description=0;
			description++;


			parent = strchr(description, ' ');
			if (!parent)
				continue;
			*parent=0;
			parent++;

			announce_resource(decode(uri), decode(description), decode(parent));
		}

		/* Test is done */
		if (line[0] == 'F') {
			char *testID;
			testID = &line[2];


			finish_test(decode(testID));
		}

		/* Test is still going, show the progress */
		if (line[0] == 'P') {
			int p;
			p = strtoull(&line[2], NULL, 10);
			report_testrun_progress(p);
		}

		free(line);
	}
	ret = pclose(file);
	if (ret){
		sprintf(buffer, "Plugin %s exited with return value %d. "
			"This probably is a bug in the plugin test itself.",
			filename, ret);
		non_fatal_bug(buffer);
	}
}


/* Go through a directory and execute all .so, .exe, .sh
 * and .pl plugins */
void execute_plugins(char *directory)
{
	DIR *dir;
	struct dirent *entry;

	dir = opendir(directory);
	if (dir==NULL) {
		printf("Plugin directory %s not found \n", directory);
		return;
	}	
	
	do {
		char buffer[PATH_MAX*2+2];
		entry = readdir(dir);
		
		if (entry && strstr(entry->d_name,".so")) {
			sprintf(buffer, "%s/%s", directory, entry->d_name);
			do_plugin(buffer);
		}
		if (entry && strstr(entry->d_name,".exe")) {
			sprintf(buffer, "%s/%s", directory, entry->d_name);
			do_exe_plugin(buffer);
		}
		if (entry && strstr(entry->d_name,".sh")) {
			sprintf(buffer, "%s/%s", directory, entry->d_name);
			do_exe_plugin(buffer);
		}
		if (entry && strstr(entry->d_name,".pl")) {
			sprintf(buffer, "%s/%s", directory, entry->d_name);
			do_exe_plugin(buffer);
		}

		
	} while (entry);	
	closedir(dir);
}
