/* 
 * Copyright (C) 2006, Intel Corporation
 * 
 * This file is part of the Linux-ready Firmware Developer Kit
 *
 * This program file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation;version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program in a file named COPYING; if not, write to the 
 * Free Software Foundation, Inc., 
 * 51 Franklin Street, Fifth Floor, 
 * Boston, MA 02110-1301 USA
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <glib.h>

#include "biostest.h"


static GList *all_tests;

GList *hardware_pokes;
GList *interactive_tests;

char *current_test;

/* find a test in 'all_tests' list, return the test if found, 
 * and NULL if not.  */
struct major_test *find_test(char *testID)
{
	struct major_test *test;
	GList *list;

	if (testID==NULL) {
		testID = current_test;
		return NULL;
	}

	list = g_list_first(all_tests);
	while (list) {
		test = (struct major_test *)list->data;
		if (test && test->ID!=NULL && strcmp(test->ID, testID)==0)
			return test;
		list = g_list_next(list);
	}
	return NULL;
}

/* after the plugin file is executed, 'start_test' is called
 * if we're at the beginning of analysing the plugin's output. */
void start_test(char *testID, char *name, char *description)
{
	struct major_test *test;
	struct detailed_result *res;

	/* first check if we're re-registering a test */
	test = find_test(testID);
	if (test) {
		/* yes we are: must be interactive, lets not do any ui things */
		current_test = test->ID;	
		return;
	}

	test = malloc(sizeof(struct major_test));
	assert(test!=NULL);

	memset(test, 0, sizeof(struct major_test));

	/* copy info of current test into 'test' */
	if (testID==NULL) { 
		testID="";
	}
	test->ID = strdup(testID);
	test->name = strdup(name);

	current_test = test->ID;

	/* add this test to the list of all_tests tested so far */
	all_tests = g_list_append(all_tests, test);

	/* if all non-interactive tests have finished, there isn't a central UI anymore */
	if (!finished_tests) {
		report_testrun_progress(0);
		report_testrun_ui(name);
	}
	if (description) {
		res = malloc(sizeof(struct detailed_result));
		assert(res!=NULL);
		memset(res, 0, sizeof(struct detailed_result));
		res->details = strdup(description);
		test->description = res;
	}

}



void report_result(char *testID, int result, char *summary, char *detailtext, char *devuri)
{
	struct major_test *test;
	struct detailed_result *details;

	if (summary==NULL)
		return;

	if (testID==NULL) 
		testID = current_test;

	test = find_test(testID);
	if (test==NULL) {
		char buf[4096];
		sprintf(buf, "report_result called on non-existing test %s\n", testID);
		non_fatal_bug(buf);
		return;
	}	
	details = malloc(sizeof(struct detailed_result));
	assert(details!=NULL);
	memset(details, 0, sizeof(struct detailed_result));

	details->result = result;
	details->summary = strdup(summary);
	if (detailtext)
		details->details = strdup(detailtext);
	if (devuri)
		details->devuri = strdup(devuri);

	test->details = g_list_append(test->details, details);
	
}

void finish_test(char *testID)
{
	struct major_test *test;
	struct detailed_result *result;
	GList *list;

	int max_result = PASS;

	report_testrun_progress(100);

	/* first check for any dmesg spewage we need to report */
	check_dmesg_updates(NULL);

	test = find_test(testID);
	if (test==NULL) {
		char buf[4096];
		sprintf(buf, "finish_test called on non-existing test %s\n", testID);
		non_fatal_bug(buf);
		return;
	}	
	list = g_list_first(test->details);
	while (list) {
		result = (struct detailed_result *)list->data;
		if (result->result > max_result)
			max_result = result->result;

		list = g_list_next(list);
	}

	test->result = max_result;
	sort_results();
}



static gint result_compare(gconstpointer a, gconstpointer b)
{
        struct major_test *testA = (struct major_test *) a;
        struct major_test *testB = (struct major_test *) b;

	if (testA->result < testB->result)
		return 1;

	if (testA->result > testB->result)
		return -1;

	return 0;
}


void sort_results(void)
{
	all_tests = g_list_sort(all_tests, result_compare);
}


void dump_results(void)
{
	dump_xml(all_tests, "/var/log/firmwarekit/results.xml");
	dump_text(all_tests, "/var/log/firmwarekit/results.txt");
	dump_xml_resources("/var/log/firmwarekit/resources.xml");
	show_results_ui(all_tests);
}
void dump_xml_results(void)
{
	dump_xml(all_tests, "/var/log/firmwarekit/results.xml");
	dump_xml_resources("/var/log/firmwarekit/resources.xml");
}

void show_results(void)
{
 	show_results_ui(all_tests);
}

void register_hw_poke(char *name, voidvoidfunc *func)
{
	struct interactive_test *test;
	test = malloc(sizeof(struct interactive_test));
	assert(test!=NULL);
	memset(test, 0, sizeof(struct interactive_test));
	test->shortname = strdup(name);
	test->execute = func;
	hardware_pokes = g_list_append(hardware_pokes, test);
}

void register_interactive_test(char *name, voidvoidfunc *func)
{
	struct interactive_test *test;
	test = malloc(sizeof(struct interactive_test));
	assert(test!=NULL);
	memset(test, 0, sizeof(struct interactive_test));
	test->shortname = strdup(name);
	test->execute = func;
	interactive_tests = g_list_append(interactive_tests, test);
}
